/************************************************************************/
/*                                                                      */
/*                      RC5 Remote Receiver                             */
/*                                                                      */
/*              Author: Peter Dannegger                                 */
/*                      danni@specs.de                                  */
/*                                                                      */
/************************************************************************/

#include <avr/interrupt.h>

#include "rc5-pd.h"

rc5_t rc5;

uint8_t  rc5_addr;
uint16_t volatile  rc5_frame = 0xffff;

#ifdef RC5_OWN_TIMER

#if defined (__AVR_ATmega8__)
#   define RC5_TCCR     TCCR0
#   define RC5_TIMSK    TIMSK

#elif defined (__AVR_ATmega88__)    \
    || defined (__AVR_ATmega168__)  \
    || defined (__AVR_ATmega328P__)
#   define RC5_TCCR     TCCR0B
#   define RC5_TIMSK    TIMSK0
#else
#   error
#endif

// Setzt bei Empfang rc5_frame (Rohdaten)
SIGNAL (SIG_OVERFLOW0)
{
    static uint8_t     rc5_bit;                // bit value
    static uint8_t     rc5_time;               // count bit time
    static uint16_t    rc5_tmp;                // shift bits in

    uint16_t tmp = rc5_tmp;               // for faster access

    TCNT0 = -(RC5CYC/256);

    if (++rc5_time > PULSE_MAX)
    {         // count pulse time
        if (!(tmp & 0x4000) && (tmp & 0x2000))   // only if 14 bits received
        {
            if (rc5_frame >= 0x8000)
                rc5_frame = tmp;
        }
        tmp = 0;
    }

    if ((rc5_bit ^ N_TO_PIN (PORT_IR)) & ID_TO_BIT (PORT_IR))      // change detect
    {
        rc5_bit = ~rc5_bit;             // 0x00 -> 0xFF -> 0x00

        if (rc5_time < PULSE_MIN)          // too short
            tmp = 0;

        if (!tmp || rc5_time > PULSE_1_2)     // start or long pulse time
        {
            if (!(tmp & 0x4000))         // not to many bits
                tmp <<= 1;              // shift
            if (!(rc5_bit & ID_TO_BIT (PORT_IR)))        // inverted bit
                tmp |= 1;               // insert new bit
            rc5_time = 0;             // count next pulse time
        }
    }

    rc5_tmp = tmp;
}
#endif // RC5_OWN_TIMER

// Decodiert rc5_frame nach rc5
// 0: Nicht ok, nix empfangen
// 1: Frame mit passender Adresse empfangen
uint8_t rc5_decode (void)
{
    uint8_t sreg = SREG;
    cli();

    union
    {
        uint8_t b[2];
        uint16_t w;
    } bw = {.w = rc5_frame};

    rc5_frame = 0xffff;
    SREG = sreg;

    if (bw.w == 0xffff)
        return 0;

    uint8_t code = bw.b[0] & 0b111111;
    bw.w <<= 2;
    int8_t flip = 0;
    if (bw.b[1] & (1 << 5))
        flip = 1;
    uint8_t addr = bw.b[1] & 0b11111;
    if (addr == rc5_addr || rc5_addr == RC5_ALL)
    {
        rc5.code = code;
        rc5.addr = addr;
        rc5.flip = flip;

        return 1;
    }
    return 0;
}

void rc5_init (uint8_t addr)
{
#ifdef RC5_OWN_TIMER
    // Timer aktivieren mit Prescale = 256
    RC5_TCCR = (1 << CS02) | (0 << CS00);

    {
        // Timer-IRQ aktivieren (Atomic!)
        uint8_t sreg = SREG;
        cli();
        RC5_TIMSK |= 1 << TOIE0;
        SREG = sreg;
    }
#endif // RC5_OWN_TIMER

    MAKE_IN (PORT_IR);
    SET     (PORT_IR);

    rc5.flip = -1;
    rc5_addr = addr;
}

