/*
    This file is part of COMMON AVR UTILS.

    COMMON AVR UTILS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    COMMON AVR UTILS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include <avr/io.h>
#include <avr/interrupt.h>

#include "reload.h"
#include "attrib.h"
#include "uart.h"

#if defined (__AVR_ATmega88__)     \
    || defined (__AVR_ATmega168__) \
    || defined (__AVR_ATmega328P__)
#define UCSRA UCSR0A
#define UCSRB UCSR0B
#define UCSRC UCSR0C
#define UDR   UDR0
#define UDRIE UDRIE0
#define RXEN  RXEN0
#define TXEN  TXEN0
#define RXC   RXC0
#define TXC   TXC0
#define TXCIE TXCIE0
#define RXCIE RXCIE0
#define UCSZ1 UCSZ01
#define UCSZ0 UCSZ00
#define UBRRL UBRR0L
#define UBRRH UBRR0H
#define U2X   U2X0
#define UDRE  UDRE0
#define SIG_UART_DATA  SIG_USART_DATA
#define SIG_UART_TRANS SIG_USART_TRANS
#define SIG_UART_RECV  SIG_USART_RECV
#endif

#if !defined (U2X_BIT)
#define U2X_BIT 1
#endif // U2X_BIT

#if !defined (BAUDRATE)
#error Define BAUDRATE
#endif // BAUDRATE

unsigned char volatile uart_transmits_p;

void uart_init(void)
{
    unsigned short ubrr = -.6 + F_CPU/((16L-8*U2X_BIT)*BAUDRATE);

    UBRRH = ubrr>>8;
    UBRRL = ubrr;

    UCSRA = (U2X_BIT << U2X) | (1 << TXC);

    // Enable UART Receiver, Transmitter
    // Data mode 8N1, asynchronous

    UCSRB = 
#ifdef UART_OUTFIFO    
    (1 << TXCIE) |
#endif // UART_OUTFIFO    
#ifdef UART_INFIFO    
    (1 << RXCIE) | 
#endif // UART_INFIFO    
    (1 << RXEN) | (1 << TXEN);

#if defined (URSEL)
    UCSRC = (1 << URSEL) | (1 << UCSZ1) | (1 << UCSZ0);
#else
    UCSRC = (1 << UCSZ1) | (1 << UCSZ0);
#endif
    do
        UDR;
    while (UCSRA & (1 << RXC));

    uart_transmits_p = 0;
}

#ifndef UART_INFIFO
char uart_getc(void)
{
    while (!(UCSRA & (1 << RXC)))
        ;
    return (char) UDR;
}

int uart_getc_nowait(void)
{
    return (UCSRA & (1 << RXC)) ? 0xff & UDR : -1;
}
#else // UART_INFIFO

static uint8_t buf_infifo[UART_INFIFO] NOINIT;
fifo_t uart_infifo;

void CONSTRUCTOR uart_init_infifo(void)
{
    fifo_init (& uart_infifo, buf_infifo, sizeof (buf_infifo));
}

char uart_getc(void)
{
    while (!uart_infifo.count)
        ;
    return fifo_get (&uart_infifo);
}

int uart_getc_nowait(void)
{
    return fifo_get (&uart_infifo);
}

SIGNAL(SIG_UART_RECV)
{
    fifo_t * infifo = &uart_infifo;
    
    RELOAD ("y", infifo);
    
    if (infifo->count < sizeof (buf_infifo))
        _inline_fifo_put_nocli (infifo, UDR);
}


#endif // UART_INFIFO

/////////////////////////////////////////////////////////////////

#ifndef UART_OUTFIFO
void uart_putc (const char c)
{
    while (!(UCSRA & (1 << UDRE)))
        ;
    UDR = c;
}

//EMPTY_INTERRUPT (SIG_UART_TRANS)

#else // UART_OUTFIFO

#if (UART_OUTFIFO < 2)
#error UART_OUTFIFO (FIFO buffer size) too small
#endif

static uint8_t buf_outfifo[UART_OUTFIFO] NOINIT;
fifo_t uart_outfifo;

void CONSTRUCTOR uart_init_outfifo(void)
{
    fifo_init (& uart_outfifo, buf_outfifo, sizeof (buf_outfifo));
}

void uart_putc (const char c)
{
#ifdef UART_WAIT_OUTFIFO
    // If IRQs are disabled, return if FIFO is full
    // otherwise, wait for fifo
    while (uart_outfifo.count >= (uint8_t) (uart_outfifo.size-2))
        asm volatile ("brid 0f");
#else // UART_WAIT_OUFIFO
    if (uart_outfifo.count < (uint8_t) (uart_outfifo.size-1))
#endif // UART_WAIT_OUFIFO
    {
        fifo_put (&uart_outfifo, c);
        UCSRB |= (1 << UDRIE);
        uart_transmits_p = !0;
    }
    asm volatile ("0:");
}

SIGNAL (SIG_UART_TRANS)
{
    uart_transmits_p = !! uart_outfifo.count;
}

SIGNAL (SIG_UART_DATA)
{
    fifo_t * of = & uart_outfifo;

    RELOAD ("y", of);

    if (0 == of->count)
    {
        UCSRB &= ~(1 << UDRIE);
        return;
    }

    UDR = _inline_fifo_get_nocli (of);
}

#endif // UART_OUTFIFO

