/*
    This file is part of MORPHEUS.

    MORPHEUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    MORPHEUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include "compat/interrupt.h"
#include "compat/pgmspace.h"
#include "attrib.h"

#include "main.h"
#include "wire.h"
#include "sintab.h"

#include "frame.h"
#include "rc5-codes.h"

#include "schoner.h"

#define NCUR 5

typedef struct schoner_t
{
    uint8_t cur;
    uint16_t i, i2, ity;
    uint16_t inc;
    uint8_t show_conf;
    int8_t v[NCUR];
} schoner_t;

schoner_t schoner = 
{
    .show_conf = 1,
    .inc    = 1, 
    .v = {2, -1, 3, 0, 0}
};

#include "morse.h"

static uint8_t 
onKey_schoner (schoner_t * t, uint8_t code)
{
    uint8_t cur = t->cur;
    if (code == RC5_LEFT)         cur--;
    else if (code == RC5_RIGHT)   cur++;
    else if (code == RC5_UP2)     t->v[cur]++;
    else if (code == RC5_DOWN2)   t->v[cur]--;
    
    if (((int8_t) cur) >= NCUR) cur -= NCUR;
    if (((int8_t) cur) < 0) cur += NCUR;
    
    t->cur = cur;
    t->v[cur] = ((t->v[cur] + 8) & 15)-8;
    
    cli();
    int16_t inc = t->inc;
    sei();
    
    if (code == RC5_11)        inc--;
    else if (code == RC5_12)   inc++;
    else if (code == RC5_TV)   inc+=10;
    else if (code == RC5_PR)   inc-=6;
    
    if (inc <= 0)  inc = 1;
    if (inc >=720) inc = 719;

    cli();
    t->inc = inc;
    t->i = 0;
    sei();
    
    if (9 == code)
        t->show_conf = !t->show_conf;
    
    return 1;
}

static inline uint16_t 
norm_720 (uint16_t un)
{
    int16_t n = un;
    
    do n -= 720;
    while (n >= 0);
    
    do n += 720;
    while (n < 0);
    
    return n;
}

frac8_t get_sin8 (uint16_t i)
{
    frac8_t si = _i_get_sin8 (norm_720(i));
    
    frac8_t f = FRAC8 (0.4); // 1/2 sqrt(2)
    
    si = fmul8_8 (si, f);
    
    return si;
}

frac8_t get_cos8 (uint16_t i)
{
    frac8_t co = _i_get_cos8 (norm_720(i));
    
    frac8_t f = FRAC8 (0.4); // 1/2 sqrt(2)
    
    co = fmul8_8 (co, f);
    
    return co;
}

static uint8_t  // NOINLINE
lissajous (schoner_t * t)
{
//    RELOAD ("y", t);
    
    uint16_t i = t->i;
    uint16_t it = t->i2;

    if (i == 0)
        beam_skip();
    i += t->inc;
    uint16_t ity = t->ity;

    if (i >= 720)
    {
        i -= 720;
        if (!i)
        {
        
            it += 2;
            ity += t->v[4];
            if (ity >= 720)
                ity = 0;
            t->ity = ity;

            if (it >= 720)
                it = 0;

            t->i2 = it;
        }
    }
    t->i = i;
    
    int8_t x = get_cos8 (i*t->v[0] + it) + get_cos8 (i*t->v[1] + it);
    int8_t y = get_sin8 (i*t->v[2] + ity) + get_sin8 (i*t->v[3] + ity);
    
    x += 0x80;
    y += 0x80;
 
    beam_xy (x, y);
    
    return !i;
}

static void 
onPixel_schoner (schoner_t * t, void * tf UNUSED)
{
    static char c;
    
    if (t->show_conf
        && !IS_FLAG (FRAME_LOCKED))
    {

        vecfont_t *vf = &vecfont;

        vf->text         = (char*) & framebuf;
        vf->proportional = 0;
        vf->zoom         = 1;
        vf->text_in_flash= 0;
        vf->start.x = 95;
        vf->start.y = 15;
        
        c = 0;
    }
    
    FLAGS |= FRAME_LOCKED;
    
    if (!c 
        && t->show_conf)
    {
        if (!vecfont_draw_letters (& vecfont))
            return;
            
        c = 1;
    }
    else
    {
        if (lissajous (t))
            FLAGS &= ~FRAME_LOCKED;
    }
}

#include <string.h>

static void 
onRedraw_schoner (schoner_t * t, void * tf UNUSED)
{
    // Leer
    char * s = (char*) & framebuf;
    
    for (uint8_t i=0; i < NCUR; i++)
    {
        int8_t v = t->v[i];
        char c = ' ';
        if (v < 0)
            v = -v, c = '-';
        *s++ = c;
        *s++ = '0'+v;
        if (i == t->cur)
        {
            *s++ = '\b';
            *s++ = '_';
        }
        *s++ = ' ';
    }
    
    *s++ = 0;
}

static void 
onEnter_schoner (schoner_t * t UNUSED)
{
}

const menuitem_t menuitem_schoner PROGMEM =
{
    .onPixel     = onPixel_schoner,
    .onEnter     = onEnter_schoner,
    .onKey       = onKey_schoner,
    .onRedraw    = onRedraw_schoner,
    .redrawOn    = REDRAW_KEY,
    .bufsize     = 0,
    .data        = & schoner
};
