/*
    This file is part of MORPHEUS.

    MORPHEUS is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License (LGPL)
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    MORPHEUS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License at <http://www.gnu.org/licenses/>
    for more details.

    (C) Georg-Johann Lay <georgjohann@web.de>
*/

#include <string.h> // memcpy

#include "compat/pgmspace.h"
#include "attrib.h"

#include "rc5-codes.h"

#include "uhr.h"
#include "frame.h"
#include "time.h"

#define UHR_RADIUS      127
// Lnge der Marken (Minuten, 5er-Minuten, Stunde (12 Uhr))

#define UHR_MARK        5
#define UHR_MARK_5      9
#define UHR_MARK_60     13

static uint8_t NOINLINE next_pixel_zifferblatt (uint8_t first, uint8_t sec);

// Sekundenzeiger
static const bh_polypoint_t uhr_sekunden_zeiger[] PROGMEM = 
{
    {{  0,  10}, BL_SKIP },
    {{  0, 110}, BL_DRAW_START | BL_LIST_END  }
};
    
// Minutenzeiger
static const bh_polypoint_t uhr_minuten_zeiger[] PROGMEM = 
{
    {{ -5,  10}, BL_SKIP },
    {{  5,  10}, BL_DRAW_START },
    {{ 10,  40}, BL_DRAW_START },
    {{  0, 100}, BL_DRAW_START },
    {{-10,  40}, BL_DRAW_START },
    {{ -5,  10}, BL_DRAW_START | BL_LIST_END }
};
    
// Stundenzeiger
static const bh_polypoint_t uhr_stunden_zeiger[] PROGMEM = 
{
    {{ -6,  10}, BL_SKIP },
    {{  6,  10}, BL_DRAW_END },
    {{ 14,  35}, BL_DRAW_END },
    {{  0,  85}, BL_DRAW_END },
    {{-14,  35}, BL_DRAW_END },
    {{ -6,  10},  BL_DRAW_END | BL_LIST_END }
};

enum
{
    UHR_ZEIG_SEKUNDE,
    UHR_ZEIG_MINUTE,
    UHR_ZEIG_STUNDE
};

static const bh_polypoint_t * uhr_sprites[] PROGMEM =
{
    [UHR_ZEIG_SEKUNDE] = uhr_sekunden_zeiger,
    [UHR_ZEIG_MINUTE]  = uhr_minuten_zeiger,
    [UHR_ZEIG_STUNDE]  = uhr_stunden_zeiger
};

static uint8_t 
onKey_uhr (uhr_t * u, uint8_t code)
{
    int16_t offset = 0;
    
    if (code == RC5_LEFT)         offset = -60;
    else if (code == RC5_RIGHT)   offset =  60;
    else if (code == RC5_UP2)     offset = 600ul;
    else if (code == RC5_DOWN2)   offset = -600l;
    else 
        return 0;
  
    u->offset += offset;    
    
    return 1;
}

static void 
onEnter_uhr (uhr_t * u)
{
    memset (u, 0, sizeof (*u));
    
    u->first = 1;
    // u->lineno = 0;
    // u->offset = 0;
    u->bpoly.linedraw = & u->bline;
    u->bpoly.sprites  = uhr_sprites;
    u->zeiger[UHR_ZEIG_SEKUNDE].sprite_id = UHR_ZEIG_SEKUNDE; 
    u->zeiger[UHR_ZEIG_MINUTE].sprite_id  = UHR_ZEIG_MINUTE;
    u->zeiger[UHR_ZEIG_STUNDE].sprite_id  = UHR_ZEIG_STUNDE;
}

// onRedraw
static void
onRedraw_uhr (uhr_t * u, void * v UNUSED)
{
    {
        short offset = u->offset;
        short inc = 1;
        if (offset < 0)
            inc = -1;
        else if (offset == 0)
            inc = 0;
        u->offset -= inc;    
        timeid_add (&time, inc);
    }
 
    bh_sprite_t * zeiger = u->zeiger;
    
    zeiger[UHR_ZEIG_SEKUNDE].rot = (uint16_t) 12 * time.second;
    
    // Sekunden % 5
    uint8_t sec5 = -1;
    int8_t sec = time.second;
    
    do 
        sec5++, sec -= 5; 
    while (sec >= 0);
    
    zeiger[UHR_ZEIG_MINUTE].rot = (uint16_t) 12 * time.minute + sec5;
    zeiger[UHR_ZEIG_STUNDE].rot = (uint16_t) 60 * time.hour + time.minute;
}

static uint8_t 
next_pixel_zeiger (uint8_t first)
{
    if (first)
    {
        // Soeben wurde der Puffer (mglicherweise)
        // aktualisiert und/oder der letzte Durchlauf der
        // Anzeige war fertig. Anzeige-Zyklus beginnt von neuem.
        bh_initpoly (& framebuf.uhr.bpoly, framebuf.uhr.zeiger, 3);
    }
    
    return bh_drawpoly (& framebuf.uhr.bpoly);
}

static void 
onPixel_uhr (uhr_t * u, void * _ UNUSED)
{
    FLAGS |= FRAME_LOCKED;
    
    uint8_t first  = u->first;
    uint8_t lineno = u->lineno;
    
    uint8_t done = (lineno < 61)
        ? next_pixel_zifferblatt (first, lineno)
        : next_pixel_zeiger (first);
    
    first = done;
    
    if (done)
        if (++lineno == 61+1) // 61 Striche + 1 Satz Sprites
        {
            lineno = 0;
            FLAGS &= ~FRAME_LOCKED;
        }
    
    u->first  = first;
    u->lineno = lineno;
}


static inline uint8_t
fmul16_16 (frac16_t a, frac16_t b)
{
    int16_t c = _i_fmul16_16 (a, b);
    
    return F16_TO_VAL (c);
}

static vf_upixel_t
frotate16_8 (uint16_t radius, point16_t sincos)
{
    uint8_t x = fmul16_16 (radius, sincos.x);
    uint8_t y = fmul16_16 (radius, sincos.y);
    return (vf_upixel_t) {x, y};
}

// Ziffernblatt pixeln
uint8_t
next_pixel_zifferblatt (uint8_t first, uint8_t sec)
{
    if (first)
    {
        // Erster Aufruf: Linie initialisieren
        uint16_t radius = UHR_RADIUS;
        uint8_t  len;
        
        uint16_t index = (uint16_t) 12 * sec;
        
        // Lnge der Markierung bestimmen
        if (sec == 0 || sec == 60)
            len = UHR_MARK_60;
        else
        {
            len = UHR_MARK;
            
            // sec % 5 bestimmen
            int8_t sec5 = sec;
            do sec5 -= 15; while (sec5 >= 0);
            do sec5 +=  5; while (sec5 <  0);
            
            if (0 == sec5)
                len = UHR_MARK_5;
        }
        
        uint16_t radius2 = radius - len;

        point16_t sincos = _i_get_sincos (index);
        
        radius  <<= 8;
        radius2 <<= 8;

        vf_upixel_t start = frotate16_8 (radius,  sincos);
        vf_upixel_t end   = frotate16_8 (radius2, sincos);
        
        if (sec ==  0)   start.x++, end.x++;
        if (sec == 60)   start.x--, end.x--;
        
        bh_initline (& framebuf.uhr.bline, 
                     (void*) 0 /* beam_xy*/, 
                     start, end, BL_DRAW_START | BL_DRAW_END);
        
        beam_skip();
    }
    
    return bh_drawline (& framebuf.uhr.bline);
}

const menuitem_t menuitem_uhr PROGMEM = 
{
    .onPixel        = onPixel_uhr,
    .onRedraw       = onRedraw_uhr,
    .onKey          = onKey_uhr,
    .onEnter        = onEnter_uhr,
    .redrawOn       = REDRAW_1_MILLI_SECOND,
    .bufsize        = 0,
    .data           = & framebuf.uhr
};
